#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "util.h"

#define MAXKEY 10000

int     **map;
BEPLONG  *key;

static
int searchKey( BEPLONG *genome, int G, int nKey )
{
     int i, j, flag;

     for ( i = 0; i < nKey; i++ ){
    
	  flag = 0;
	  
	  for ( j = 0; j < G; j++ ){

	       if ( genome[j] == key[G*i +j] ) flag++;
	       else break;
	  }

	  if ( flag == G ) return i;
     }

     return -1; //--- Key is not stored.
}

static
int createKey( BigSimulator2D *SIM )
{
     int nKey, MaxKey;
     int i, j, k, kk, x, y;
     int minX, maxX, minY, maxY;

     BEPLONG *temp;
     
     MaxKey = MAXKEY; // Default number of key array size;
     nKey   = 1;
     minX   = SIM->minX;
     maxX   = SIM->maxX;
     minY   = SIM->minY;
     maxY   = SIM->maxY;
     
     /* ------------------------ */
   
     for ( y = minY; y <= maxY; y++ ){
        for ( x = minX; x <= maxX; x++ ){
            if( isEmpty(x,y,SIM) ){
                continue;
            }
           
	       i = coordinate2index( x, SIM );
	       j = coordinate2index( y, SIM );

	       k = searchKey( SIM->genomes[i][j], SIM->G, nKey ) ;
          
          
          

	       if ( k < 0 ){ //--- New key
		    if ( nKey == MaxKey ){//--- Reallocate key table
			 MaxKey += MAXKEY;
			 temp = (BEPLONG *)realloc( key, MaxKey*SIM->G*sizeof(BEPLONG) );
			 if ( !temp ){
			      writeMessage("BEP_err","Fail to allocate temporary memory for creating key map.\n");
			      writeMessage("BEP_err","BEP is aborted by error.\n");
			      abort();
			 }
			 //free(key);
			 key = temp;
		    }
		    
		    for ( kk = 0; kk < SIM->G; kk++ ) key[SIM->G*nKey +kk] = SIM->genomes[i][j][kk];
		    map[ x - minX ][ y - minY ] = nKey;
            nKey++;
	       }
	       else{
		    map[ x - minX ][ y - minY ] = k;
	       }
	       
	  }
     }
     //fprintf(stderr,"BEP_dbg, (print:createKey) MaxKey = %d\n",MaxKey);
     
     return nKey;     
}

static
void writeTM( BigSimulator2D *SIM )
{
    FILE *fp;
    char  name[256];
    int   x, y;
    int minX, maxX, minY, maxY;
     
     /* ------------ File open ------------ */
     sprintf(name,"%s.tm",SIM->prefix);
     if ( (fp = fopen(name,"w")) == NULL ){
	  fprintf(stderr,"BEP_err: Fail to open %s file.\n",name);
	  exit(0);
     }
    minX   = SIM->minX;
    maxX   = SIM->maxX;
    minY   = SIM->minY;
    maxY   = SIM->maxY;
    
    
    for(y = maxY; y>=minY; y--){
        for(x = minX; x<=maxX; x++){
            if( isEmpty(x,y,SIM) ){
                fprintf(fp,"%d\t",0);
            }else{
                if( isStem(x,y,SIM) ){
                    fprintf(fp,"%d\t", map[ x - minX ][ y - minY ]+1);
                }else{
                    fprintf(fp,"-%d\t", map[ x - minX ][ y - minY ]+1);
                }
            }
	  }
	  fprintf(fp,"\n");
    }

     fclose(fp);
}

static
void writeMUT( int nKey, BigSimulator2D *SIM )
{
     FILE *fp;
     char  name[256];
     char  buff[MAXBINARY+1] = "";
     //char  *buff;
     int   i, j, k, G, len;

     //buff = (char *)malloc( (MAXBINARY+1)*sizeof(char) );
     //if (!buff){
     //  writeMessage("BEP_err","Fail to allocate char buffer memory in writing MUT file.\n");
     //  abort();
     //}
     
     /* ------------ File open ------------ */
     sprintf(name,"%s.mut",SIM->prefix);
     if ( (fp = fopen(name,"w")) == NULL ){
	  fprintf(stderr,"BEP_err: Fail to open %s file.\n",name);
	  exit(0);
     }

     G = SIM->G;
     for ( i = 0; i < nKey; i++ ){

	  fprintf(fp,"%d\t",i+1);

	  for ( j = 0; j < G-1; j++ ){
	       //fprintf(stderr,"(MUT) i, j, key buff = %d %d %ld ",i,j,key[G*i+j]);
	       decimal2binary( key[G*i+j], buff );
	       //fprintf(stderr,"%s\n",buff);
	       //fprintf(fp,"%s",buff);
	       for ( k = 0; k < MAXBINARY; k++){
               fprintf(fp, "%c\t", buff[k]);
	       }
	  }
    len = (SIM->genomeSize) - (G-1)*MAXBINARY;
      decimal2binary( key[G*i+(G-1)], buff );
      for ( k = 0; k < len; k++){
             fprintf(fp, "%c\t",buff[k]);
     }
         
	  fprintf(fp,"\n");
     }

     fclose(fp);
     //free(buff);
}

static
void writePRM( BigSimulator2D *SIM )
{
     FILE *fp;
     char  name[256];
     
     /* ------------ File open ------------ */
     sprintf(name,"%s.prm",SIM->prefix);
     if ( (fp = fopen(name,"w")) == NULL ){
	  fprintf(stderr,"BEP_err: Fail to open %s file.\n",name);
	  exit(0);
     }

    
     /*fprintf(fp,"mutationRate                   = %e\n",SIM->mutationRate);
     fprintf(fp,"growthRate                     = %e\n",SIM->growthRate);
     fprintf(fp,"deathRate                      = %e\n",SIM->deathRate);
     fprintf(fp,"deathRateForNonStem            = %e\n",SIM->deathRateForNonStem);
     fprintf(fp,"symmetricReplicationProbablity = %e\n",SIM->symmetricReplicationProbablity);
     fprintf(fp,"fitnessIncrease                = %e\n",SIM->fitnessIncrease);

     fprintf(fp,"maxTime               = %d\n",SIM->maxTime);
     fprintf(fp,"maxPopulationSize     = %d\n",SIM->maxPopulationSize);
     fprintf(fp,"genomeSize            = %d\n",SIM->genomeSize);
     fprintf(fp,"driverSize            = %d\n",SIM->driverSize);
     fprintf(fp,"essensialGeneSize     = %d\n",SIM->essensialGeneSize);
     fprintf(fp,"initialPopulationSize = %d\n",SIM->initialPopulationSize);

     fprintf(fp,"populationSize        = %d\n",SIM->populationSize);
     fprintf(fp,"time                  = %d\n",SIM->time);
     fprintf(fp,"{min,max} X           = %d %d\n",SIM->minX,SIM->maxX);
     fprintf(fp,"{min,max} Y           = %d %d\n",SIM->minY,SIM->maxY);
     fprintf(fp,"limXY                 = %d\n",SIM->limXY);
     fprintf(fp,"G                     = %d\n",SIM->G);*/
    
    fprintf(fp,"mutationRate\t%e\n",SIM->mutationRate);
    fprintf(fp,"growthRate\t%e\n",SIM->growthRate);
    fprintf(fp,"deathRate\t%e\n",SIM->deathRate);
    fprintf(fp,"deathRateForNonStem\t%e\n",SIM->deathRateForNonStem);
    fprintf(fp,"symmetricReplicationProbablity\t%e\n",SIM->symmetricReplicationProbablity);
    fprintf(fp,"fitnessIncrease\t%e\n",SIM->fitnessIncrease);
    
    fprintf(fp,"maxTime\t%d\n",SIM->maxTime);
    fprintf(fp,"maxPopulationSize\t%d\n",SIM->maxPopulationSize);
    fprintf(fp,"genomeSize\t%d\n",SIM->genomeSize);
    fprintf(fp,"driverSize\t%d\n",SIM->driverSize);
    fprintf(fp,"essensialGeneSize\t%d\n",SIM->essensialGeneSize);
    fprintf(fp,"initialPopulationSize\t%d\n",SIM->initialPopulationSize);
    
    fprintf(fp,"populationSize\t%d\n",SIM->populationSize);
    fprintf(fp,"time\t%d\n",SIM->time);
    fprintf(fp,"minX\t%d\n",SIM->minX);
    fprintf(fp,"maxX\t%d\n",SIM->maxX);
    fprintf(fp,"minY\t%d\n",SIM->minY);
    fprintf(fp,"maxY\t%d\n",SIM->maxY);
    fprintf(fp,"limXY\t%d\n",SIM->limXY);
    fprintf(fp,"G\t%d\n",SIM->G);
    

     fclose(fp);
}

void printResults( BigSimulator2D *SIM )
{
     //FILE *os3; // .prm

     int       nKey;     
     int       i, x, y, *m;
     int       minX, maxX, minY, maxY;

     writeMessage("BEP_msg","Writing results to files...\n");

     /* ------------ Memory allocation ------------ */
     minX   = SIM->minX;
     maxX   = SIM->maxX;
     minY   = SIM->minY;
     maxY   = SIM->maxY;

     key = mallocLong( MAXKEY*SIM->G );
     for ( i = 0; i < SIM->G; i++ ) key[i] = 0; // Key for vacant

     y = maxY - minY +1;
     x = maxX - minX +1;
     map = (int **)malloc( x*sizeof(int*) );
     if ( !map ){
	  writeMessage("BEP_err","Fail to allocate temporary memory for output.\n");
	  writeMessage("BEP_err","BEP is aborted by error.\n");
	  abort();
     }
     m = (int *)malloc( x*y*sizeof(int) );
     if ( !m ){
	  writeMessage("BEP_err","Fail to allocate temporary memory for output.\n");
	  writeMessage("BEP_err","BEP is aborted by error.\n");
	  abort();
     }
     for ( i = 0; i < x*y; i++ ) m[i] = -1;
     for ( i = 0; i < x; i++ ) map[i] = m + y*i;
     /* ------------------------ */

     nKey = createKey( SIM );
     
     writeTM( SIM );
     writeMUT( nKey, SIM );
     writePRM( SIM );
     
     free(key);
     free(map[0]);
     free(map);

     writeMessage("BEP_msg","Writing results to files...done.\n");
}
